# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with Speos feature: Source."""

from __future__ import annotations

from difflib import SequenceMatcher
from typing import List, Mapping, Optional, Union
import uuid

from ansys.speos.core import (
    project as project,
    proto_message_utils as proto_message_utils,
)
import ansys.speos.core.body as body
import ansys.speos.core.face as face
from ansys.speos.core.geo_ref import GeoRef
from ansys.speos.core.intensity import Intensity
from ansys.speos.core.kernel.client import SpeosClient
from ansys.speos.core.kernel.scene import ProtoScene
from ansys.speos.core.kernel.source_template import ProtoSourceTemplate
from ansys.speos.core.spectrum import Spectrum


class BaseSource:
    """
    Super Class for all sources.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project in which source shall be created.
    name : str
        Name of the source.
    description : str
        Description of the source.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    source_instance : ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance, optional
        Source instance to provide if the feature does not have to be created from scratch
        By default, ``None``, means that the feature is created from scratch by default.

    Notes
    -----
    This is a Super class, **Do not instantiate this class yourself**
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
    ) -> None:
        self._project = project
        self._name = name
        self._unique_id = None
        self.source_template_link = None
        """Link object for the source template in database."""

        if metadata is None:
            metadata = {}

        if source_instance is None:
            # Create local SourceTemplate
            self._source_template = ProtoSourceTemplate(
                name=name, description=description, metadata=metadata
            )

            # Create local SourceInstance
            self._source_instance = ProtoScene.SourceInstance(
                name=name, description=description, metadata=metadata
            )
        else:
            self._unique_id = source_instance.metadata["UniqueId"]
            self.source_template_link = self._project.client[source_instance.source_guid]
            self._reset()

    class _Spectrum:
        def __init__(
            self,
            speos_client: SpeosClient,
            name: str,
            message_to_complete: Union[
                ProtoSourceTemplate.RayFile,
                ProtoSourceTemplate.Surface,
                ProtoSourceTemplate.Luminaire,
            ],
            spectrum_guid: str = "",
        ) -> None:
            self._message_to_complete = message_to_complete
            if spectrum_guid != "":
                self._spectrum = Spectrum(
                    speos_client=speos_client,
                    name=name + ".Spectrum",
                    key=spectrum_guid,
                )
            else:
                self._spectrum = Spectrum(speos_client=speos_client, name=name + ".Spectrum")

            self._no_spectrum = None  # None means never committed, or deleted
            self._no_spectrum_local = False

        def __str__(self) -> str:
            if self._no_spectrum is None:
                if self._no_spectrum_local is False:
                    return str(self._spectrum)
            else:
                if self._no_spectrum is False:
                    return str(self._spectrum)
            return ""

        def _commit(self) -> BaseSource._Spectrum:
            if not self._no_spectrum_local:
                self._spectrum.commit()
                self._message_to_complete.spectrum_guid = self._spectrum.spectrum_link.key
                self._no_spectrum = self._no_spectrum_local
            return self

        def _reset(self) -> BaseSource._Spectrum:
            self._spectrum.reset()
            if self._no_spectrum is not None:
                self._no_spectrum_local = self._no_spectrum
            return self

        def _delete(self) -> BaseSource._Spectrum:
            self._no_spectrum = None
            return self

    def _to_dict(self) -> dict:
        out_dict = {}

        # SourceInstance (= source guid + source properties)
        if self._project.scene_link and self._unique_id is not None:
            scene_data = self._project.scene_link.get()
            src_inst = next(
                (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if src_inst is not None:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client, message=src_inst
                )
            else:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self._source_instance,
                )
        else:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._project.client, message=self._source_instance
            )

        if "source" not in out_dict.keys():
            # SourceTemplate
            if self.source_template_link is None:
                out_dict["source"] = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self._source_template,
                )
            else:
                out_dict["source"] = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self.source_template_link.get(),
                )

        # # handle spectrum & intensity
        # if self._type is not None:
        #     self._type._to_dict(dict_to_complete=out_dict)

        proto_message_utils._replace_properties(json_dict=out_dict)

        return out_dict

    def get(self, key: str = "") -> list[tuple[str, dict]]:
        """Get dictionary corresponding to the project - read only.

        Parameters
        ----------
        key: str

        Returns
        -------
        str | dict
        """
        if key == "":
            return self._to_dict()
        info = proto_message_utils._value_finder_key_startswith(dict_var=self._to_dict(), key=key)
        content = list(info)
        if len(content) != 0:
            content.sort(
                key=lambda x: SequenceMatcher(None, x[0], key).ratio(),
                reverse=True,
            )
            return content[0][1]
        info = proto_message_utils._flatten_dict(dict_var=self._to_dict())
        print("Used key: {} not found in key list: {}.".format(key, info.keys()))

    def __str__(self) -> str:
        """Return the string representation of the source."""
        out_str = ""
        if self._project.scene_link and self._unique_id is not None:
            scene_data = self._project.scene_link.get()
            src_inst = next(
                (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if src_inst is None:
                out_str += "local: "
        else:
            out_str += "local: "

        out_str += proto_message_utils.dict_to_str(dict=self._to_dict())
        return out_str

    def _commit(self) -> BaseSource:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        # The _unique_id will help to find correct item in the scene.sources:
        # the list of SourceInstance
        if self._unique_id is None:
            self._unique_id = str(uuid.uuid4())
            self._source_instance.metadata["UniqueId"] = self._unique_id

        # Save or Update the source template (depending on if it was already saved before)
        if self.source_template_link is None:
            self.source_template_link = self._project.client.source_templates().create(
                message=self._source_template
            )
            self._source_instance.source_guid = self.source_template_link.key
        elif self.source_template_link.get() != self._source_template:
            self.source_template_link.set(
                data=self._source_template
            )  # Only update if template has changed

        # Update the scene with the source instance
        if self._project.scene_link:
            update_scene = True
            scene_data = self._project.scene_link.get()  # retrieve scene data

            # Look if an element corresponds to the _unique_id
            src_inst = next(
                (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if src_inst is not None:
                if src_inst != self._source_instance:
                    src_inst.CopyFrom(self._source_instance)  # if yes, just replace
                else:
                    update_scene = False
            else:
                scene_data.sources.append(
                    self._source_instance
                )  # if no, just add it to the list of sources

            if update_scene:  # Update scene only if instance has changed
                self._project.scene_link.set(data=scene_data)  # update scene data

        return self

    def _reset(self) -> BaseSource:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        # Reset source template
        if self.source_template_link is not None:
            self._source_template = self.source_template_link.get()

        # Reset source instance
        if self._project.scene_link is not None:
            scene_data = self._project.scene_link.get()  # retrieve scene data
            # Look if an element corresponds to the _unique_id
            src_inst = next(
                (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if src_inst is not None:
                self._source_instance = src_inst
        return self

    def _delete(self) -> BaseSource:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        # This allows to clean-managed object contained in _luminaire, _rayfile, etc..
        # Like Spectrum, IntensityTemplate

        # Delete the source template
        if self.source_template_link is not None:
            self.source_template_link.delete()
            self.source_template_link = None

        # Reset then the source_guid (as the source template was deleted just above)
        self._source_instance.source_guid = ""

        # Remove the source from the scene
        scene_data = self._project.scene_link.get()  # retrieve scene data
        src_inst = next(
            (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
            None,
        )
        if src_inst is not None:
            scene_data.sources.remove(src_inst)
            self._project.scene_link.set(data=scene_data)  # update scene data

        # Reset the _unique_id
        self._unique_id = None
        self._source_instance.metadata.pop("UniqueId")
        return self

    def _fill(self, src_inst):
        self._unique_id = src_inst.metadata["UniqueId"]
        self._source_instance = src_inst
        self.source_template_link = self._project.client[src_inst.source_guid]
        self._reset()

    def commit(self) -> BaseSource:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        self._spectrum._commit()
        self._commit()
        return self

    def reset(self) -> BaseSource:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        self._spectrum._reset()
        self._reset()
        return self

    def delete(self) -> BaseSource:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        self._spectrum._delete()
        self._delete()
        return self


class SourceLuminaire(BaseSource):
    """LuminaireSource.

    By default, a flux from intensity file is chosen, with an incandescent spectrum.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    default_values : bool
        Uses default values when True.
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )

        self._spectrum = self._Spectrum(
            speos_client=self._project.client,
            name=name,
            message_to_complete=self._source_template.luminaire,
            spectrum_guid=self._source_template.luminaire.spectrum_guid,
        )

        if default_values:
            # Default values
            self.set_flux_from_intensity_file().set_spectrum().set_incandescent()
            self.set_axis_system()

    def set_flux_from_intensity_file(self) -> SourceLuminaire:
        """Take flux from intensity file provided.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        self._source_template.luminaire.flux_from_intensity_file.SetInParent()
        return self

    def set_flux_luminous(self, value: float = 683) -> SourceLuminaire:
        """Set luminous flux.

        Parameters
        ----------
        value : float
            Luminous flux in lumens.
            By default, ``683.0``.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        self._source_template.luminaire.luminous_flux.luminous_value = value
        return self

    def set_flux_radiant(self, value: float = 1) -> SourceLuminaire:
        """Set radiant flux.

        Parameters
        ----------
        value : float
            Radiant flux in watts.
            By default, ``1.0``.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        self._source_template.luminaire.radiant_flux.radiant_value = value
        return self

    def set_intensity_file_uri(self, uri: str) -> SourceLuminaire:
        """Set intensity file.

        Parameters
        ----------
        uri : str
            IES or EULUMDAT format file uri.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        self._source_template.luminaire.intensity_file_uri = uri
        return self

    def set_spectrum(self) -> Spectrum:
        """Set spectrum.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum.
        """
        if self._spectrum._message_to_complete is not self._source_template.luminaire:
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._spectrum._message_to_complete = self._source_template.luminaire
        return self._spectrum._spectrum

    def set_axis_system(self, axis_system: Optional[List[float]] = None) -> SourceLuminaire:
        """Set the position of the source.

        Parameters
        ----------
        axis_system : Optional[List[float]]
            Position of the source [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
            By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        if axis_system is None:
            axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
        self._source_instance.luminaire_properties.axis_system[:] = axis_system
        return self


class SourceRayFile(BaseSource):
    """RayFile Source.

    By default, flux and spectrum from ray file are selected.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    default_values : bool
        Uses default values when True.
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )
        self._client = self._project.client

        spectrum_guid = ""
        if self._source_template.rayfile.HasField("spectrum_guid"):
            spectrum_guid = self._source_template.rayfile.spectrum_guid
        self._spectrum = self._Spectrum(
            speos_client=self._client,
            name=name,
            message_to_complete=self._source_template.rayfile,
            spectrum_guid=spectrum_guid,
        )
        if spectrum_guid == "":
            self.set_spectrum_from_ray_file()

        self._name = name

        if default_values:
            # Default values
            self.set_flux_from_ray_file().set_spectrum_from_ray_file()
            self.set_axis_system()

    def set_ray_file_uri(self, uri: str) -> SourceRayFile:
        """Set ray file.

        Parameters
        ----------
        uri : str
            Rayfile format file uri (.ray or .tm25ray files expected).

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.ray_file_uri = uri
        return self

    def set_flux_from_ray_file(self) -> SourceRayFile:
        """Take flux from ray file provided.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.flux_from_ray_file.SetInParent()
        return self

    def set_flux_luminous(self, value: float = 683) -> SourceRayFile:
        """Set luminous flux.

        Parameters
        ----------
        value : float
            Luminous flux in lumens.
            By default, ``683.0``.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.luminous_flux.luminous_value = value
        return self

    def set_flux_radiant(self, value: float = 1) -> SourceRayFile:
        """Set radiant flux.

        Parameters
        ----------
        value : float
            Radiant flux in watts.
            By default, ``1.0``.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.radiant_flux.radiant_value = value
        return self

    def set_spectrum_from_ray_file(self) -> SourceRayFile:
        """Take spectrum from ray file provided.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.spectrum_from_ray_file.SetInParent()
        self._spectrum._no_spectrum_local = True
        return self

    def set_spectrum(self) -> Spectrum:
        """Set spectrum of the Source.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum.
        """
        if self._source_template.rayfile.HasField("spectrum_from_ray_file"):
            guid = ""
            if self._spectrum._spectrum.spectrum_link is not None:
                guid = self._spectrum._spectrum.spectrum_link.key
            self._source_template.rayfile.spectrum_guid = guid

        if self._spectrum._message_to_complete is not self._source_template.rayfile:
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._spectrum._message_to_complete = self._source_template.rayfile

        self._spectrum._no_spectrum_local = False
        return self._spectrum._spectrum

    def set_axis_system(self, axis_system: Optional[List[float]] = None) -> SourceRayFile:
        """Set position of the source.

        Parameters
        ----------
        axis_system : Optional[List[float]]
            Position of the source [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
            By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile Source.
        """
        if axis_system is None:
            axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
        self._source_instance.rayfile_properties.axis_system[:] = axis_system
        return self

    def set_exit_geometries(self, exit_geometries: Optional[List[GeoRef]] = None) -> SourceRayFile:
        """Set exit geometries.

        Parameters
        ----------
        exit_geometries : List[ansys.speos.core.geo_ref.GeoRef]
            Exit Geometries that will use this rayfile source.
            By default, ``[]``.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile Source.
        """
        if not exit_geometries:
            self._source_instance.rayfile_properties.ClearField("exit_geometries")
        else:
            self._source_instance.rayfile_properties.exit_geometries.geo_paths[:] = [
                gr.to_native_link() for gr in exit_geometries
            ]

        return self


class SourceSurface(BaseSource):
    """Type of Source : Surface.

    By default, a luminous flux and existence constant are chosen. With a monochromatic spectrum,
    and lambertian intensity (cos with N = 1).

    Parameters
    ----------
    speos_client : ansys.speos.core.kernel.client.SpeosClient
        The Speos instance client.
    name : str
        Name of the source feature.
    surface : ansys.api.speos.source.v1.source_pb2.SourceTemplate.Surface
        Surface source to complete.
    surface_props : ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance.SurfaceProperties
        Surface source properties to complete.
    default_values : bool
        Uses default values when True.
    """

    class ExitanceVariable:
        """Type of surface source existence : existence variable.

        Parameters
        ----------
        exitance_variable : ansys.api.speos.source.v1.source_pb2.SourceTemplate.Surface.
        ExitanceVariable
            Existence variable to complete.
        exitance_variable_props : ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance.
        SurfaceProperties.ExitanceVariableProperties
            Existence variable properties to complete.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_exitance_variable method available in
        Source classes.
        """

        def __init__(
            self,
            exitance_variable,
            exitance_variable_props,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "ExitanceVariable class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._exitance_variable = exitance_variable
            self._exitance_variable_props = exitance_variable_props

            if default_values:
                # Default values
                self._exitance_variable.SetInParent()
                self.set_axis_plane()

        def set_xmp_file_uri(self, uri: str) -> SourceSurface.ExitanceVariable:
            """Set existence xmp file.

            Parameters
            ----------
            uri : str
                XMP file describing existence.

            Returns
            -------
            ansys.speos.core.source.SourceSurface.ExitanceVariable
                ExitanceVariable of surface source.
            """
            self._exitance_variable.exitance_xmp_file_uri = uri
            return self

        def set_axis_plane(
            self, axis_plane: Optional[List[float]] = None
        ) -> SourceSurface.ExitanceVariable:
            """Set position of the existence map.

            Parameters
            ----------
            axis_plane : Optional[List[float]]
                Position of the existence map [Ox Oy Oz Xx Xy Xz Yx Yy Yz].
                By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0]``.

            Returns
            -------
            ansys.speos.core.source.SourceSurface.ExitanceVariable
                ExitanceVariable of surface Source.
            """
            if axis_plane is None:
                axis_plane = [0, 0, 0, 1, 0, 0, 0, 1, 0]
            self._exitance_variable_props.axis_plane[:] = axis_plane
            return self

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )
        self._speos_client = self._project.client
        self._name = name

        spectrum_guid = ""
        if self._source_template.surface.HasField("spectrum_guid"):
            spectrum_guid = self._source_template.surface.spectrum_guid
        self._spectrum = self._Spectrum(
            speos_client=self._speos_client,
            name=name,
            message_to_complete=self._source_template.surface,
            spectrum_guid=spectrum_guid,
        )

        self._intensity = Intensity(
            speos_client=self._speos_client,
            name=name + ".Intensity",
            intensity_props_to_complete=self._source_instance.surface_properties.intensity_properties,
            key=self._source_template.surface.intensity_guid,
        )

        # Attribute gathering more complex existence type
        self._exitance_type = None

        if default_values:
            # Default values
            self.set_flux_luminous().set_exitance_constant(geometries=[]).set_intensity()
            self.set_spectrum()

    def set_flux_from_intensity_file(self) -> SourceSurface:
        """Take flux from intensity file provided.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.flux_from_intensity_file.SetInParent()
        return self

    def set_flux_luminous(self, value: float = 683) -> SourceSurface:
        """Set luminous flux.

        Parameters
        ----------
        value : float
            Luminous flux in lumens.
            By default, ``683.0``.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.luminous_flux.luminous_value = value
        return self

    def set_flux_radiant(self, value: float = 1) -> SourceSurface:
        """Set radiant flux.

        Parameters
        ----------
        value : float
            Radiant flux in watts.
            By default, ``1.0``.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.radiant_flux.radiant_value = value
        return self

    def set_flux_luminous_intensity(self, value: float = 5) -> SourceSurface:
        """Set luminous intensity flux.

        Parameters
        ----------
        value : float
            Luminous intensity in candelas.
            By default, ``5.0``.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.luminous_intensity_flux.luminous_intensity_value = value
        return self

    def set_intensity(self) -> Intensity:
        """Set intensity.

        Returns
        -------
        ansys.speos.core.intensity.Intensity
            Intensity.
        """
        if (
            self._intensity._intensity_properties
            is not self._source_instance.surface_properties.intensity_properties
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._intensity._intensity_properties = (
                self._source_instance.surface_properties.intensity_properties
            )

        return self._intensity

    def set_exitance_constant(
        self, geometries: List[tuple[Union[GeoRef, face.Face, body.Body], bool]]
    ) -> SourceSurface:
        """Set existence constant.

        Parameters
        ----------
        geometries : List[tuple[ansys.speos.core.geo_ref.GeoRef, bool]]
            List of (face, reverseNormal).

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._exitance_type = None

        self._source_template.surface.exitance_constant.SetInParent()
        self._source_instance.surface_properties.exitance_constant_properties.ClearField(
            "geo_paths"
        )
        if geometries != []:
            geo_paths = []
            for gr, reverse_normal in geometries:
                if isinstance(gr, GeoRef):
                    geo_paths.append(
                        ProtoScene.GeoPath(
                            geo_path=gr.to_native_link(), reverse_normal=reverse_normal
                        )
                    )
                elif isinstance(gr, (face.Face, body.Body)):
                    geo_paths.append(
                        ProtoScene.GeoPath(
                            geo_path=gr.geo_path.to_native_link(), reverse_normal=reverse_normal
                        )
                    )
                else:
                    msg = f"Type {type(gr)} is not supported as Surface Source geometry input."
                    raise TypeError(msg)
            self._source_instance.surface_properties.exitance_constant_properties.geo_paths.extend(
                geo_paths
            )
        return self

    def set_exitance_variable(self) -> SourceSurface.ExitanceVariable:
        """Set existence variable, taken from XMP map.

        Returns
        -------
        ansys.speos.core.source.SourceSurface.ExitanceVariable
            ExitanceVariable of surface source.
        """
        if self._exitance_type is None and self._source_template.surface.HasField(
            "exitance_variable"
        ):
            # Happens in case of project created via load of speos file
            self._exitance_type = SourceSurface.ExitanceVariable(
                exitance_variable=self._source_template.surface.exitance_variable,
                exitance_variable_props=self._source_instance.surface_properties.exitance_variable_properties,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._exitance_type, SourceSurface.ExitanceVariable):
            # if the _exitance_type is not ExitanceVariable then we create a new type.
            self._exitance_type = SourceSurface.ExitanceVariable(
                exitance_variable=self._source_template.surface.exitance_variable,
                exitance_variable_props=self._source_instance.surface_properties.exitance_variable_properties,
                stable_ctr=True,
            )
        elif (
            self._exitance_type._exitance_variable
            is not self._source_template.surface.exitance_variable
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._exitance_type._exitance_variable = self._source_template.surface.exitance_variable
            self._exitance_type._exitance_variable_props = (
                self._source_instance.surface_properties.exitance_variable_properties
            )
        return self._exitance_type

    def set_spectrum_from_xmp_file(self) -> SourceSurface:
        """Take spectrum from xmp file provided.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.spectrum_from_xmp_file.SetInParent()
        self._spectrum._no_spectrum_local = True
        return self

    def set_spectrum(self) -> Spectrum:
        """Set spectrum of the Source.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum.
        """
        if self._source_template.surface.HasField("spectrum_from_xmp_file"):
            guid = ""
            if self._spectrum._spectrum.spectrum_link is not None:
                guid = self._spectrum._spectrum.spectrum_link.key
            self._source_template.surface.spectrum_guid = guid

        if self._spectrum._message_to_complete is not self._source_template.surface:
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._spectrum._message_to_complete = self._source_template.surface

        self._spectrum._no_spectrum_local = False
        return self._spectrum._spectrum

    def commit(self) -> SourceSurface:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Source feature.
        """
        # intensity
        self._intensity.commit()
        self._source_template.surface.intensity_guid = self._intensity.intensity_template_link.key

        # spectrum & source
        super().commit()
        return self

    def reset(self) -> SourceSurface:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Source feature.
        """
        self._intensity.reset()
        # spectrum & source
        super().reset()
        return self

    def delete(self) -> SourceSurface:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Source feature.
        """
        # Currently we don't perform delete in cascade,
        # so deleting a surface source does not delete the intensity template used
        # self._intensity.delete()

        # spectrum & source
        super().delete()
        return self



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with Speos feature: Spectrum."""

from __future__ import annotations

from typing import List, Mapping, Optional
import warnings

from ansys.speos.core.kernel.client import SpeosClient
from ansys.speos.core.kernel.proto_message_utils import protobuf_message_to_dict
from ansys.speos.core.kernel.spectrum import ProtoSpectrum
from ansys.speos.core.proto_message_utils import dict_to_str


class Spectrum:
    """Speos feature : Spectrum.

    By default, a monochromatic spectrum is created.

    Parameters
    ----------
    speos_client : ansys.speos.core.kernel.client.SpeosClient
        The Speos instance client.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    key : str
        Creation from an SpectrumLink key

    Attributes
    ----------
    spectrum_link : ansys.speos.core.kernel.spectrum.SpectrumLink
        Link object for the spectrum in database.
    """

    def __init__(
        self,
        speos_client: SpeosClient,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        key: str = "",
    ) -> None:
        self._client = speos_client
        self.spectrum_link = None
        """Link object for the spectrum in database."""

        if metadata is None:
            metadata = {}

        if key == "":
            # Create Spectrum
            self._spectrum = ProtoSpectrum(name=name, description=description, metadata=metadata)

            # Default value
            self.set_monochromatic()  # By default will be monochromatic
        else:
            # Retrieve Spectrum
            self.spectrum_link = speos_client[key]
            self._spectrum = self.spectrum_link.get()

    def set_monochromatic(self, wavelength: float = 555.0) -> Spectrum:
        """Set the spectrum as monochromatic.

        Parameters
        ----------
        wavelength : float
            Wavelength of the spectrum, in nm.
            By default, ``555.0``.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.monochromatic.wavelength = wavelength
        return self

    def set_blackbody(self, temperature: float = 2856) -> Spectrum:
        """Set the spectrum as blackbody.

        Parameters
        ----------
        temperature : float
            Temperature of the blackbody, in K.
            By default, ``2856``.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.blackbody.temperature = temperature
        return self

    def set_sampled(self, wavelengths: List[float], values: List[float]) -> Spectrum:
        """Set the spectrum as sampled.

        Parameters
        ----------
        wavelengths : List[float]
            List of wavelengths, in nm
        values : List[float]
            List of values, expected from 0. to 100. in %

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.sampled.wavelengths[:] = wavelengths
        self._spectrum.sampled.values[:] = values
        return self

    def set_library(self, file_uri: str) -> Spectrum:
        """Set the spectrum as library.

        Parameters
        ----------
        file_uri : str
            uri of the spectrum file.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.library.file_uri = file_uri
        return self

    def set_incandescent(self) -> Spectrum:
        """Set the spectrum as incandescent (predefined spectrum).

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.predefined.incandescent.SetInParent()
        return self

    def set_warmwhitefluorescent(self) -> Spectrum:
        """Set the spectrum as warmwhitefluorescent (predefined spectrum).

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.predefined.warmwhitefluorescent.SetInParent()
        return self

    def set_daylightfluorescent(self) -> Spectrum:
        """Set the spectrum as daylightfluorescent (predefined spectrum).

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.predefined.daylightfluorescent.SetInParent()
        return self

    def set_whiteLED(self) -> Spectrum:
        """Set the spectrum as white led (predefined spectrum).

        .. deprecated:: 0.2.2
            `set_whiteLed` will be removed with 0.3.0
            `set_white_led` shall be used to comply with PEP8 naming convention

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        warnings.warn(
            "`set_whiteLED` is deprecated. Use `set_white_led` method instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.set_white_led()

    def set_white_led(self) -> Spectrum:
        """Set the spectrum as white led (predefined spectrum).

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.predefined.whiteLED.SetInParent()
        return self

    def set_halogen(self) -> Spectrum:
        """Set the spectrum as halogen (predefined spectrum).

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.predefined.halogen.SetInParent()
        return self

    def set_metalhalide(self) -> Spectrum:
        """Set the spectrum as metalhalide (predefined spectrum).

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.predefined.metalhalide.SetInParent()
        return self

    def set_highpressuresodium(self) -> Spectrum:
        """Set the spectrum as highpressuresodium (predefined spectrum).

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        self._spectrum.predefined.highpressuresodium.SetInParent()
        return self

    def _to_dict(self) -> dict:
        if self.spectrum_link is None:
            return protobuf_message_to_dict(self._spectrum)
        else:
            return protobuf_message_to_dict(message=self.spectrum_link.get())

    def __str__(self) -> str:
        """Return the string representation of the spectrum."""
        out_str = ""
        if self.spectrum_link is None:
            out_str += "local: "
        out_str += dict_to_str(self._to_dict())
        return out_str

    def commit(self) -> Spectrum:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        if self.spectrum_link is None:
            self.spectrum_link = self._client.spectrums().create(message=self._spectrum)
        elif self.spectrum_link.get() != self._spectrum:
            self.spectrum_link.set(data=self._spectrum)  # Only update if data has changed

        return self

    def reset(self) -> Spectrum:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        if self.spectrum_link is not None:
            self._spectrum = self.spectrum_link.get()
        return self

    def delete(self) -> Spectrum:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum feature.
        """
        if self.spectrum_link is not None:
            self.spectrum_link.delete()
            self.spectrum_link = None

        return self



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""PySpeos Kernel module gathers low-level interactions and internal operations of the project."""

from ansys.speos.core.kernel.body import BodyLink, ProtoBody
from ansys.speos.core.kernel.client import SpeosClient
from ansys.speos.core.kernel.face import FaceLink, ProtoFace
from ansys.speos.core.kernel.intensity_template import (
    IntensityTemplateLink,
    ProtoIntensityTemplate,
)
from ansys.speos.core.kernel.job import JobLink, ProtoJob
from ansys.speos.core.kernel.part import PartLink, ProtoPart
from ansys.speos.core.kernel.proto_message_utils import (
    protobuf_message_to_dict,
    protobuf_message_to_str,
)
from ansys.speos.core.kernel.scene import ProtoScene, SceneLink
from ansys.speos.core.kernel.sensor_template import (
    ProtoSensorTemplate,
    SensorTemplateLink,
)
from ansys.speos.core.kernel.simulation_template import (
    ProtoSimulationTemplate,
    SimulationTemplateLink,
)
from ansys.speos.core.kernel.sop_template import (
    ProtoSOPTemplate,
    SOPTemplateLink,
)
from ansys.speos.core.kernel.source_template import (
    ProtoSourceTemplate,
    SourceTemplateLink,
)
from ansys.speos.core.kernel.spectrum import ProtoSpectrum, SpectrumLink
from ansys.speos.core.kernel.vop_template import (
    ProtoVOPTemplate,
    VOPTemplateLink,
)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a wrapped abstraction of the gRPC proto API definition and stubs."""

from typing import List

from ansys.api.speos.source.v1 import (
    source_pb2 as messages,
    source_pb2_grpc as service,
)
from ansys.speos.core.kernel.crud import CrudItem, CrudStub
from ansys.speos.core.kernel.proto_message_utils import protobuf_message_to_str

ProtoSourceTemplate = messages.SourceTemplate
"""SourceTemplate protobuf class : ansys.api.speos.source.v1.source_pb2.SourceTemplate"""
ProtoSourceTemplate.__str__ = lambda self: protobuf_message_to_str(self)


class SourceTemplateLink(CrudItem):
    """Link object for a source template in database.

    Parameters
    ----------
    db : ansys.speos.core.kernel.source_template.SourceTemplateStub
        Database to link to.
    key : str
        Key of the source template in the database.
    """

    def __init__(self, db, key: str):
        super().__init__(db, key)
        self._actions_stub = db._actions_stub

    def __str__(self) -> str:
        """Return the string representation of the source template."""
        return str(self.get())

    def get(self) -> ProtoSourceTemplate:
        """Get the datamodel from database.

        Returns
        -------
        source_template.SourceTemplate
            Source template datamodel.
        """
        return self._stub.read(self)

    def set(self, data: ProtoSourceTemplate) -> None:
        """Change datamodel in database.

        Parameters
        ----------
        data : source_template.SourceTemplate
            New source template datamodel.
        """
        self._stub.update(self, data)

    def delete(self) -> None:
        """Remove datamodel from database."""
        self._stub.delete(self)

    # Actions
    def get_ray_file_info(self) -> messages.GetRayFileInfo_Response:
        """
        Retrieve information about ray file source.

        Returns
        -------
        ansys.api.speos.source.v1.source_pb2.GetRayFileInfo_Response
            Information about ray file source, like flux value.
        """
        return self._actions_stub.GetRayFileInfo(messages.GetRayFileInfo_Request(guid=self.key))


class SourceTemplateStub(CrudStub):
    """
    Database interactions for source templates.

    Parameters
    ----------
    channel : grpc.Channel
        Channel to use for the stub.

    Examples
    --------
    The best way to get a SourceTemplateStub is to retrieve it from SpeosClient via
    source_templates() method. Like in the following example:

    >>> from ansys.speos.core.speos import Speos
    >>> speos = Speos(host="localhost", port=50098)
    >>> src_t_db = speos.client.source_templates()

    """

    def __init__(self, channel):
        super().__init__(stub=service.SourceTemplatesManagerStub(channel=channel))
        self._actions_stub = service.SourceTemplateActionsStub(channel=channel)

    def create(self, message: ProtoSourceTemplate) -> SourceTemplateLink:
        """Create a new entry.

        Parameters
        ----------
        message : source_template.SourceTemplate
            Datamodel for the new entry.

        Returns
        -------
        ansys.speos.core.kernel.source_template.SourceTemplateLink
            Link object created.
        """
        resp = CrudStub.create(self, messages.Create_Request(source_template=message))
        return SourceTemplateLink(self, resp.guid)

    def read(self, ref: SourceTemplateLink) -> ProtoSourceTemplate:
        """Get an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.source_template.SourceTemplateLink
            Link object to read.

        Returns
        -------
        source_template.SourceTemplate
            Datamodel of the entry.
        """
        if not ref.stub == self:
            raise ValueError("SourceTemplateLink is not on current database")
        resp = CrudStub.read(self, messages.Read_Request(guid=ref.key))
        return resp.source_template

    def update(self, ref: SourceTemplateLink, data: ProtoSourceTemplate):
        """Change an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.source_template.SourceTemplateLink
            Link object to update.
        data : source_template.SourceTemplate
            New datamodel for the entry.
        """
        if not ref.stub == self:
            raise ValueError("SourceTemplateLink is not on current database")
        CrudStub.update(self, messages.Update_Request(guid=ref.key, source_template=data))

    def delete(self, ref: SourceTemplateLink) -> None:
        """Remove an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.source_template.SourceTemplateLink
            Link object to delete.
        """
        if not ref.stub == self:
            raise ValueError("SourceTemplateLink is not on current database")
        CrudStub.delete(self, messages.Delete_Request(guid=ref.key))

    def list(self) -> List[SourceTemplateLink]:
        """List existing entries.

        Returns
        -------
        List[ansys.speos.core.kernel.source_template.SourceTemplateLink]
            Link objects.
        """
        guids = CrudStub.list(self, messages.List_Request()).guids
        return list(map(lambda x: SourceTemplateLink(self, x), guids))



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a wrapped abstraction of the gRPC proto API definition and stubs."""

from typing import Iterator, List

from ansys.api.speos.results.v1.ray_path_pb2 import RayPath
from ansys.api.speos.scene.v2 import (
    scene_pb2 as messages,
    scene_pb2_grpc as service,
)
from ansys.speos.core.kernel.crud import CrudItem, CrudStub
from ansys.speos.core.kernel.proto_message_utils import protobuf_message_to_str

ProtoScene = messages.Scene
"""Scene protobuf class : ansys.api.speos.scene.v2.scene_pb2.Scene"""
ProtoScene.__str__ = lambda self: protobuf_message_to_str(self)
ProtoScene.MaterialInstance.__str__ = lambda self: protobuf_message_to_str(self)
ProtoScene.SceneInstance.__str__ = lambda self: protobuf_message_to_str(self)
ProtoScene.SourceInstance.__str__ = lambda self: protobuf_message_to_str(self)
ProtoScene.SensorInstance.__str__ = lambda self: protobuf_message_to_str(self)
ProtoScene.SimulationInstance.__str__ = lambda self: protobuf_message_to_str(self)


class SceneLink(CrudItem):
    """
    Link object for a scene in database.

    Parameters
    ----------
    db : ansys.speos.core.kernel.scene.SceneStub
        Database to link to.
    key : str
        Key of the scene in the database.

    Examples
    --------
    >>> from ansys.speos.core.speos import Speos
    >>> from ansys.speos.core.kernel.scene import ProtoScene
    >>> speos = Speos(host="localhost", port=50098)
    >>> sce_db = speos.client.scenes()
    >>> sce_link = sce_db.create(message=ProtoScene(name="Empty_Scene"))

    """

    def __init__(self, db, key: str):
        super().__init__(db, key)
        self._actions_stub = db._actions_stub

    def __str__(self) -> str:
        """Return the string representation of the scene."""
        return str(self.get())

    def get(self) -> ProtoScene:
        """Get the datamodel from database.

        Returns
        -------
        scene.Scene
            Scene datamodel.
        """
        return self._stub.read(self)

    def set(self, data: ProtoScene) -> None:
        """Change datamodel in database.

        Parameters
        ----------
        data : scene.Scene
            New scene datamodel.
        """
        self._stub.update(self, data)

    def delete(self) -> None:
        """Remove datamodel from database."""
        self._stub.delete(self)

    # Actions
    def load_file(self, file_uri: str) -> None:
        """
        Load speos file to fill the scene.

        Parameters
        ----------
        file_uri : str
            File to be loaded.
        """
        self._actions_stub.LoadFile(messages.LoadFile_Request(guid=self.key, file_uri=file_uri))

    def get_source_ray_paths(self, source_path: str, rays_nb: int = 100) -> Iterator[RayPath]:
        """
        Retrieve source ray paths.

        Parameters
        ----------
        source_path : str
            Path to the source in the Scene : "<source name>" for a specific source in the current
            scene, or "<sub-scene name>/<source name>" for a specific source in a specific sub
            scene.
        rays_nb : int, optional
            Number of rays generated by the source.
            By default, ``100``.

        Returns
        -------
        Iterator[ansys.api.speos.results.v1.ray_path_pb2.RayPath]
            Ray paths generated by the source.
        """
        for rp in self._actions_stub.GetSourceRayPaths(
            messages.GetSourceRayPaths_Request(
                guid=self.key, source_path=source_path, rays_nb=rays_nb
            )
        ):
            yield rp


class SceneStub(CrudStub):
    """
    Database interactions for scenes.

    Parameters
    ----------
    channel : grpc.Channel
        Channel to use for the stub.

    Examples
    --------
    The best way to get a SceneStub is to retrieve it from SpeosClient via scenes() method.
    Like in the following example:

    >>> from ansys.speos.core.speos import Speos
    >>> speos = Speos(host="localhost", port=50098)
    >>> sce_db = speos.client.scenes()

    """

    def __init__(self, channel):
        super().__init__(stub=service.ScenesManagerStub(channel=channel))
        self._actions_stub = service.SceneActionsStub(channel=channel)

    def create(self, message: ProtoScene = None) -> SceneLink:
        """Create a new entry.

        Parameters
        ----------
        message : scene.Scene, optional.
            Datamodel for the new entry.

        Returns
        -------
        ansys.speos.core.kernel.scene.SceneLink
            Link object created.
        """
        if message is None:
            message = ProtoScene()
        resp = CrudStub.create(self, messages.Create_Request(scene=message))
        return SceneLink(self, resp.guid)

    def read(self, ref: SceneLink) -> ProtoScene:
        """Get an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.scene.SceneLink
            Link object to read.

        Returns
        -------
        scene.Scene
            Datamodel of the entry.
        """
        if not ref.stub == self:
            raise ValueError("SceneLink is not on current database")
        resp = CrudStub.read(self, messages.Read_Request(guid=ref.key))
        return resp.scene

    def update(self, ref: SceneLink, data: ProtoScene):
        """Change an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.scene.SceneLink
            Link object to update.
        data : scene.Scene
            New datamodel for the entry.
        """
        if not ref.stub == self:
            raise ValueError("SceneLink is not on current database")
        CrudStub.update(self, messages.Update_Request(guid=ref.key, scene=data))

    def delete(self, ref: SceneLink) -> None:
        """Remove an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.scene.SceneLink
            Link object to delete.
        """
        if not ref.stub == self:
            raise ValueError("SceneLink is not on current database")
        CrudStub.delete(self, messages.Delete_Request(guid=ref.key))

    def list(self) -> List[SceneLink]:
        """List existing entries.

        Returns
        -------
        List[ansys.speos.core.kernel.scene.SceneLink]
            Link objects.
        """
        guids = CrudStub.list(self, messages.List_Request()).guids
        return list(map(lambda x: SceneLink(self, x), guids))



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Open one of the possible results generated out of the simulation."""

import os
from pathlib import Path
import tempfile
from typing import Union

import ansys.api.speos.file.v1.file_transfer as file_transfer_helper__v1
import ansys.api.speos.file.v1.file_transfer_pb2_grpc as file_transfer__v1__pb2_grpc

if os.name == "nt":
    from comtypes.client import CreateObject

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
from numpy import ndarray

from ansys.speos.core.simulation import (
    SimulationDirect,
    SimulationInteractive,
    SimulationInverse,
)


def _find_correct_result(
    simulation_feature: Union[SimulationDirect, SimulationInverse, SimulationInteractive],
    result_name: str,
    download_if_distant: bool = True,
) -> str:
    if len(simulation_feature.result_list) == 0:
        raise ValueError("Please compute the simulation feature to generate results.")

    file_path = ""

    for res in simulation_feature.result_list:
        if res.HasField("path"):
            if res.path.endswith(result_name):
                file_path = res.path
                break
        elif res.HasField("upload_response"):
            if res.upload_response.info.file_name == result_name:
                if download_if_distant:
                    file_transfer_helper__v1.download_file(
                        file_transfer_service_stub=file_transfer__v1__pb2_grpc.FileTransferServiceStub(
                            simulation_feature._project.client.channel
                        ),
                        file_uri=res.upload_response.info.uri,
                        download_location=tempfile.gettempdir(),
                    )
                    file_path = str(
                        Path(tempfile.gettempdir()) / res.upload_response.info.file_name
                    )
                else:
                    file_path = res.upload_response.info.uri
                break
    return file_path


def _display_image(img: ndarray):
    if img is not None:
        plt.imshow(img)
        plt.axis("off")  # turns off axes
        plt.axis("tight")  # gets rid of white border
        plt.axis("image")  # square up the image instead of filling the "figure" space
        plt.show()


if os.name == "nt":

    def open_result_image(
        simulation_feature: Union[SimulationDirect, SimulationInverse, SimulationInteractive],
        result_name: str,
    ) -> None:
        """Retrieve an image from a specific simulation result.

        Parameters
        ----------
        simulation_feature : ansys.speos.core.simulation.Simulation
            The simulation feature.
        result_name : str
            The result name to open as an image.
        """
        file_path = _find_correct_result(simulation_feature, result_name)
        if file_path == "":
            raise ValueError(
                "No result corresponding to "
                + result_name
                + " is found in "
                + simulation_feature._name
            )

        if file_path.endswith("xmp") or file_path.endswith("XMP"):
            dpf_instance = CreateObject("XMPViewer.Application")
            dpf_instance.OpenFile(file_path)
            res = dpf_instance.ExportXMPImage(file_path + ".png", 1)
            if res:
                _display_image(mpimg.imread(file_path + ".png"))
        elif file_path.endswith("png") or file_path.endswith("PNG"):
            _display_image(mpimg.imread(file_path))

    def open_result_in_viewer(
        simulation_feature: Union[SimulationDirect, SimulationInverse],
        result_name: str,
    ) -> None:
        """Open a specific simulation result in the suitable viewer.

        Parameters
        ----------
        simulation_feature : ansys.speos.core.simulation.Simulation
            The simulation feature.
        result_name : str
            The result name to open in a viewer.
        """
        file_path = _find_correct_result(simulation_feature, result_name)

        if file_path.endswith("xmp") or file_path.endswith("XMP"):
            dpf_instance = CreateObject("XMPViewer.Application")
            dpf_instance.OpenFile(file_path)
            dpf_instance.Show(1)
        elif file_path.endswith("hdr") or file_path.endswith("HDR"):
            dpf_instance = CreateObject("HDRIViewer.Application")
            dpf_instance.OpenFile(file_path)
            dpf_instance.Show(1)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides the ``Speos`` class."""

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from grpc import Channel

from ansys.speos.core.generic.constants import (
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEFAULT_VERSION,
    MAX_CLIENT_MESSAGE_SIZE,
)
from ansys.speos.core.kernel.client import SpeosClient

if TYPE_CHECKING:  # pragma: no cover
    from ansys.platform.instancemanagement import Instance


class Speos:
    """Allows the Speos session (client) to interact with the SpeosRPC server.

    Parameters
    ----------
    host : str, optional
        Host where the server is running.
        By default, ``ansys.speos.core.kernel.client.DEFAULT_HOST``.
    port : Union[str, int], optional
        Port number where the server is running.
        By default, ``ansys.speos.core.kernel.client.DEFAULT_PORT``.
    version : str
        The Speos server version to run, in the 3 digits format, such as "242".
        If unspecified, the version will be chosen as
        ``ansys.speos.core.kernel.client.LATEST_VERSION``.
    channel : ~grpc.Channel, optional
        gRPC channel for server communication.
        By default, ``None``.
    message_size: int
        Maximum Message size of a newly generated channel
        By default, ``MAX_CLIENT_MESSAGE_SIZE``.
    remote_instance : ansys.platform.instancemanagement.Instance
        The corresponding remote instance when the Speos Service
        is launched through PyPIM. This instance will be deleted when calling
        :func:`SpeosClient.close <ansys.speos.core.kernel.client.SpeosClient.close >`.
    timeout : Real, optional
        Timeout in seconds to achieve the connection.
        By default, 60 seconds.
    logging_level : int, optional
        The logging level to be applied to the client.
        By default, ``INFO``.
    logging_file : Optional[str, Path]
        The file to output the log, if requested. By default, ``None``.
    """

    def __init__(
        self,
        host: str = DEFAULT_HOST,
        port: Union[str, int] = DEFAULT_PORT,
        version: str = DEFAULT_VERSION,
        channel: Optional[Channel] = None,
        message_size: int = MAX_CLIENT_MESSAGE_SIZE,
        remote_instance: Optional["Instance"] = None,
        timeout: Optional[int] = 60,
        logging_level: Optional[int] = logging.INFO,
        logging_file: Optional[Union[Path, str]] = None,
        speos_install_path: Optional[Union[Path, str]] = None,
    ):
        self._client = SpeosClient(
            host=host,
            port=port,
            version=version,
            channel=channel,
            message_size=message_size,
            remote_instance=remote_instance,
            timeout=timeout,
            logging_level=logging_level,
            logging_file=logging_file,
            speos_install_path=speos_install_path,
        )

    @property
    def client(self) -> SpeosClient:
        """The ``Speos`` instance client."""
        return self._client

    def close(self) -> bool:
        """Close the channel and deletes all Speos objects from memory.

        Returns
        -------
        bool
            Information if the server instance was terminated.

        Notes
        -----
        If an instance of the Speos Service was started using
        PyPIM, this instance will be deleted.
        """
        return self.client.close()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides the ``VisualData`` class."""

from typing import TYPE_CHECKING, List

from ansys.speos.core.generic.general_methods import (
    graphics_required,
    magnitude_vector,
    normalize_vector,
)

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv


@graphics_required
class _VisualCoordinateSystem:
    """Visualization data for the coordinate system..

    By default, there is empty visualization data.

    Notes
    -----
    **Do not instantiate this class yourself**.
    """

    def __init__(self):
        import pyvista as pv

        self.__origin = [0.0, 0.0, 0.0]
        self.__x_axis = pv.Arrow(
            start=self.__origin,
            direction=[1.0, 0.0, 0.0],
            scale=10.0,
            tip_radius=0.05,
            shaft_radius=0.01,
        )
        self.__y_axis = pv.Arrow(
            start=self.__origin,
            direction=[0.0, 1.0, 0.0],
            scale=10.0,
            tip_radius=0.05,
            shaft_radius=0.01,
        )
        self.__z_axis = pv.Arrow(
            start=self.__origin,
            direction=[0.0, 0.0, 1.0],
            scale=10.0,
            tip_radius=0.05,
            shaft_radius=0.01,
        )

    @property
    def origin(self) -> List[float]:
        """Returns the origin of the coordinate system.

        Returns
        -------
        List[float]
            The origin of the coordinate system.

        """
        return self.__origin

    @origin.setter
    def origin(self, value: List[float]) -> None:
        """Set the origin of the coordinate system.

        Parameters
        ----------
        value: List[float]
            The origin of the coordinate system.

        Returns
        -------
        None
        """
        if len(value) != 3:
            raise ValueError("origin must be a List of three values.")
        self.__origin = value

    @property
    def x_axis(self) -> "pv.PolyData":
        """Returns the x-axis of the coordinate system.

        Returns
        -------
        pv.PolyData
            pyvista.PolyData the x-axis of the coordinate system.

        """
        return self.__x_axis

    @x_axis.setter
    def x_axis(self, x_vector: List[float]) -> None:
        """Set the x-axis of the coordinate system.

        Parameters
        ----------
        x_vector: List[float]
            The x-axis of the coordinate system.

        Returns
        -------
        None
        """
        import pyvista as pv

        if len(x_vector) != 3:
            raise ValueError("x_axis must be a List of three values.")
        self.__x_axis = pv.Arrow(
            start=self.__origin,
            direction=normalize_vector(vector=x_vector),
            scale=magnitude_vector(vector=x_vector),
            tip_radius=0.05,
            shaft_radius=0.01,
        )

    @property
    def y_axis(self) -> "pv.PolyData":
        """
        Returns the y-axis of the coordinate system.

        Returns
        -------
        pv.PolyData
            pyvista.PolyData the y-axis of the coordinate system.

        """
        return self.__y_axis

    @y_axis.setter
    def y_axis(self, y_vector: List[float]) -> None:
        """Set the y-axis of the coordinate system.

        Parameters
        ----------
        y_vector: List[float]
            The y-axis of the coordinate system.

        Returns
        -------
        None
        """
        import pyvista as pv

        if len(y_vector) != 3:
            raise ValueError("y_axis must be a List of three values.")
        self.__y_axis = pv.Arrow(
            start=self.__origin,
            direction=normalize_vector(vector=y_vector),
            scale=magnitude_vector(vector=y_vector),
            tip_radius=0.05,
            shaft_radius=0.01,
        )

    @property
    def z_axis(self) -> "pv.PolyData":
        """
        Returns the z-axis of the coordinate system.

        Returns
        -------
        pv.PolyData
            pyvista.PolyData the z-axis of the coordinate system.

        """
        return self.__z_axis

    @z_axis.setter
    def z_axis(self, z_vector: List[float]) -> None:
        """Set the z-axis of the coordinate system.

        Parameters
        ----------
        z_vector: List[float]
            The z-axis of the coordinate system.

        Returns
        -------
        None
        """
        import pyvista as pv

        if len(z_vector) != 3:
            raise ValueError("z_axis must be a List of three values.")
        self.__z_axis = pv.Arrow(
            start=self.__origin,
            direction=normalize_vector(vector=z_vector),
            scale=magnitude_vector(vector=z_vector),
            tip_radius=0.05,
            shaft_radius=0.01,
        )


@graphics_required
class _VisualData:
    """Visualization data for the sensor.

    By default, there is empty visualization data.

    Notes
    -----
    **Do not instantiate this class yourself**
    """

    def __init__(self, coordinate_system=True):
        import pyvista as pv

        self.__data = pv.PolyData()
        self.coordinates = _VisualCoordinateSystem() if coordinate_system else None
        self.updated = False

    @property
    def data(self) -> "pv.PolyData":
        """
        Returns the data of the sensor.

        Returns
        -------
        pv.PolyData
            The data of the surface visualization.

        """
        return self.__data

    def add_data_triangle(self, triangle_vertices: List[List[float]]) -> None:
        """
        Add surface data triangle to Visualization data.

        Parameters
        ----------
        triangle_vertices: List[List[float]]
            The vertices of the triangle.

        Returns
        -------
        None
        """
        import pyvista as pv

        if len(triangle_vertices) != 3 or any(len(vertex) != 3 for vertex in triangle_vertices):
            raise ValueError(
                "triangle_vertices is expected to be composed of 3 vertices with 3 elements each."
            )
        faces = [[3, 0, 1, 2]]
        self.__data = self.__data.append_polydata(pv.PolyData(triangle_vertices, faces))

    def add_data_rectangle(self, rectangle_vertices: List[List[float]]) -> None:
        """
        Add surface data rectangle to Visualization data.

        Parameters
        ----------
        rectangle_vertices: List[List[float]]
            The vertices of the rectangle.

        Returns
        -------
        None
        """
        import pyvista as pv

        if len(rectangle_vertices) != 3 or any(len(vertex) != 3 for vertex in rectangle_vertices):
            raise ValueError(
                "rectangle_vertices is expected to be composed of 3 vertices with 3 elements each."
            )
        self.__data = self.__data.append_polydata(pv.Rectangle(rectangle_vertices))



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides interface to link Speos Objects to Geometries."""

from __future__ import annotations

from typing import Mapping


class GeoRef:
    """Represent a CAD object."""

    def __init__(self, name: str, description: str, metadata: Mapping[str, str]):
        self.name = name
        self.description = description
        self.metadata = metadata
        return

    @staticmethod
    def from_native_link(geopath: str) -> GeoRef:
        """
        Convert a native link to a GeoRef.

        Parameters
        ----------
        geopath : str
            Geometry path.

        Returns
        -------
        GeoRef
        """
        return GeoRef("", "", {"GeoPath": geopath})

    def to_native_link(self) -> str:
        """
        Convert to a native link.

        Returns
        -------
        str
            Geometry path.
        """
        return self.metadata["GeoPath"]



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a way to interact with Speos feature: Optical Property."""

from __future__ import annotations

from difflib import SequenceMatcher
from typing import List, Mapping, Optional, Union
import uuid

import ansys.speos.core.body as body
import ansys.speos.core.face as face
from ansys.speos.core.geo_ref import GeoRef
from ansys.speos.core.kernel.scene import ProtoScene
from ansys.speos.core.kernel.sop_template import ProtoSOPTemplate
from ansys.speos.core.kernel.vop_template import ProtoVOPTemplate
import ansys.speos.core.part as part
import ansys.speos.core.project as project
import ansys.speos.core.proto_message_utils as proto_message_utils


class OptProp:
    """Speos feature: optical property.

    By default, a mirror 100% is chosen as surface optical property,
    without any volume optical property.
    By default, the optical property is applied to no geometry.

    Parameters
    ----------
    project : project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str, optional
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]], optional
        Metadata of the feature.
        By default, ``None``.
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
    ):
        self._name = name
        self._project = project
        self._unique_id = None
        self.sop_template_link = None
        """Link object for the sop template in database."""
        self.vop_template_link = None
        """Link object for the vop template in database."""

        # Create SOP template
        if metadata is None:
            metadata = {}
        self._sop_template = ProtoSOPTemplate(
            name=name + ".SOP", description=description, metadata=metadata
        )

        # Create VOP template
        self._vop_template = None

        # Create material instance
        self._material_instance = ProtoScene.MaterialInstance(
            name=name, description=description, metadata=metadata
        )

        # Default values
        self.set_surface_mirror().set_volume_none().set_geometries()

    def set_surface_mirror(self, reflectance: float = 100) -> OptProp:
        """
        Perfect specular surface.

        Parameters
        ----------
        reflectance : float
            Reflectance, expected from 0. to 100. in %.
            By default, ``100``.

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical property.
        """
        self._sop_template.mirror.reflectance = reflectance
        return self

    def set_surface_opticalpolished(self) -> OptProp:
        """
        Transparent or perfectly polished material (glass, plastic).

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical property.
        """
        self._sop_template.optical_polished.SetInParent()
        return self

    def set_surface_library(self, path: str) -> OptProp:
        r"""
        Based on surface optical properties file.

        Parameters
        ----------
        path : str
            Surface optical properties file, \*.scattering, \*.bsdf, \*.brdf, \*.coated, ...

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical property.
        """
        self._sop_template.library.sop_file_uri = path
        return self

    def set_volume_none(self) -> OptProp:
        """
        No volume optical property.

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical property.
        """
        self._vop_template = None
        return self

    def set_volume_opaque(self) -> OptProp:
        """
        Non transparent material.

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical property.
        """
        if self._vop_template is None:
            self._vop_template = ProtoVOPTemplate(
                name=self._name + ".VOP",
                description=self._sop_template.description,
                metadata=self._sop_template.metadata,
            )
        self._vop_template.opaque.SetInParent()
        return self

    def set_volume_optic(
        self,
        index: float = 1.5,
        absorption: float = 0,
        constringence: Optional[float] = None,
    ) -> OptProp:
        """
        Transparent colorless material without bulk scattering.

        Parameters
        ----------
        index : float
            Refractive index.
            By default, ``1.5``.
        absorption : float
            Absorption coefficient value. mm-1.
            By default, ``0``.
        constringence : float, optional
            Abbe number.
            By default, ``None``, means no constringence.

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical property.
        """
        if self._vop_template is None:
            self._vop_template = ProtoVOPTemplate(
                name=self._name + ".VOP",
                description=self._sop_template.description,
                metadata=self._sop_template.metadata,
            )
        self._vop_template.optic.index = index
        self._vop_template.optic.absorption = absorption
        if constringence is not None:
            self._vop_template.optic.constringence = constringence
        else:
            self._vop_template.optic.ClearField("constringence")
        return self

    # Deactivated due to a bug on SpeosRPC server side
    # def set_volume_nonhomogeneous(
    #         self,
    #         path: str,
    #         axis_system: Optional[List[float]] = None
    # ) -> OptProp:
    #    """
    #    Material with non-homogeneous refractive index.
    #
    #    Parameters
    #    ----------
    #    path : str
    #        \*.gradedmaterial file that describes the spectral variations of
    #        refractive index and absorption with the respect to position in space.
    #    axis_system : Optional[List[float]]
    #        Orientation of the non-homogeneous material [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
    #        By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.
    #
    #    Returns
    #    -------
    #    ansys.speos.core.opt_prop.OptProp
    #        Optical property.
    #    """
    #    if not axis_system:
    #        axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
    #    if self._vop_template is None:
    #        self._vop_template = VOPTemplate(
    #            name=self._name + ".VOP",
    #            description=self._sop_template.description,
    #            metadata=self._sop_template.metadata
    #        )
    #    self._vop_template.non_homogeneous.gradedmaterial_file_uri = path
    #    self._material_instance.non_homogeneous_properties.axis_system[:] = axis_system
    #    return self

    def set_volume_library(self, path: str) -> OptProp:
        r"""
        Based on \*.material file.

        Parameters
        ----------
        path : str
            \*.material file

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical property.
        """
        if self._vop_template is None:
            self._vop_template = ProtoVOPTemplate(
                name=self._name + ".VOP",
                description=self._sop_template.description,
                metadata=self._sop_template.metadata,
            )
        self._vop_template.library.material_file_uri = path
        return self

    def set_geometries(
        self,
        geometries: Optional[List[Union[GeoRef, body.Body, face.Face, part.Part.SubPart]]] = None,
    ) -> OptProp:
        """Select geometries on which the optical properties will be applied.

        Parameters
        ----------
        geometries : List[ansys.speos.core.geo_ref.GeoRef], optional
            List of geometries. Giving an empty list means "All geometries"
            By default, ``None``, means "no geometry".

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical property.
        """
        if geometries is None:
            self._material_instance.ClearField("geometries")
        else:
            geo_paths = []
            for gr in geometries:
                if isinstance(gr, GeoRef):
                    geo_paths.append(gr)
                elif isinstance(gr, (face.Face, body.Body, part.Part.SubPart)):
                    geo_paths.append(gr.geo_path)
                else:
                    msg = f"Type {type(gr)} is not supported as Optical property geometry input."
                    raise TypeError(msg)
            self._material_instance.geometries.geo_paths[:] = [
                gp.to_native_link() for gp in geo_paths
            ]
        return self

    def _to_dict(self) -> dict:
        out_dict = {}

        # MaterialInstance (= vop guid + sop guids + geometries)
        if self._project.scene_link and self._unique_id is not None:
            scene_data = self._project.scene_link.get()
            mat_inst = next(
                (x for x in scene_data.materials if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if mat_inst is not None:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client, message=mat_inst
                )
            else:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self._material_instance,
                )
        else:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._project.client,
                message=self._material_instance,
            )

        if "vop" not in out_dict.keys():
            # SensorTemplate
            if self.vop_template_link is None:
                if self._vop_template is not None:
                    out_dict["vop"] = proto_message_utils._replace_guids(
                        speos_client=self._project.client,
                        message=self._vop_template,
                    )
            else:
                out_dict["vop"] = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self.vop_template_link.get(),
                )

        if "sops" not in out_dict.keys():
            # SensorTemplate
            if self.sop_template_link is None:
                if self._sop_template is not None:
                    out_dict["sops"] = [
                        proto_message_utils._replace_guids(
                            speos_client=self._project.client,
                            message=self._sop_template,
                        )
                    ]
            else:
                out_dict["sops"] = [
                    proto_message_utils._replace_guids(
                        speos_client=self._project.client,
                        message=self.sop_template_link.get(),
                    )
                ]

        proto_message_utils._replace_properties(json_dict=out_dict)

        return out_dict

    def get(self, key: str = "") -> str | dict:
        """Get dictionary corresponding to the project - read only.

        Parameters
        ----------
        key: str

        Returns
        -------
        str | dict
        """
        if key == "":
            return self._to_dict()
        info = proto_message_utils._value_finder_key_startswith(dict_var=self._to_dict(), key=key)
        content = list(info)
        if len(content) != 0:
            content.sort(
                key=lambda x: SequenceMatcher(None, x[0], key).ratio(),
                reverse=True,
            )
            return content[0][1]
        info = proto_message_utils._flatten_dict(dict_var=self._to_dict())
        print("Used key: {} not found in key list: {}.".format(key, info.keys()))

    def __str__(self):
        """Return the string representation of the optical property."""
        out_str = ""
        # MaterialInstance (= vop guid + sop guids + geometries)
        if self._project.scene_link and self._unique_id is not None:
            scene_data = self._project.scene_link.get()
            mat_inst = next(
                (x for x in scene_data.materials if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if mat_inst is None:
                out_str += "local: "
        else:
            out_str += "local: "

        out_str += proto_message_utils.dict_to_str(dict=self._to_dict())
        return out_str

    def commit(self) -> OptProp:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            Optical Property feature.
        """
        # The _unique_id will help to find correct item in the scene.materials:
        # the list of MaterialInstance
        if self._unique_id is None:
            self._unique_id = str(uuid.uuid4())
            self._material_instance.metadata["UniqueId"] = self._unique_id

        # Save or Update the vop template (depending on if it was already saved before)
        if self.vop_template_link is None:
            if self._vop_template is not None:
                self.vop_template_link = self._project.client.vop_templates().create(
                    message=self._vop_template
                )
                self._material_instance.vop_guid = self.vop_template_link.key
        elif self.vop_template_link.get() != self._vop_template:
            self.vop_template_link.set(
                data=self._vop_template
            )  # Only update if vop template has changed

        # Save or Update the sop template (depending on if it was already saved before)
        if self.sop_template_link is None:
            if self._sop_template is not None:
                self.sop_template_link = self._project.client.sop_templates().create(
                    message=self._sop_template
                )
                self._material_instance.ClearField("sop_guids")
                self._material_instance.sop_guids.append(self.sop_template_link.key)
        elif self.sop_template_link.get() != self._sop_template:
            self.sop_template_link.set(
                data=self._sop_template
            )  # Only update if sop template has changed

        # Update the scene with the material instance
        if self._project.scene_link:
            update_scene = True
            scene_data = self._project.scene_link.get()  # retrieve scene data

            # Look if an element corresponds to the _unique_id
            mat_inst = next(
                (x for x in scene_data.materials if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if mat_inst is not None:
                if mat_inst != self._material_instance:
                    mat_inst.CopyFrom(self._material_instance)  # if yes, just replace
                else:
                    update_scene = False
            else:
                scene_data.materials.append(
                    self._material_instance
                )  # if no, just add it to the list of material instances

            if update_scene:  # Update scene only if instance has changed
                self._project.scene_link.set(data=scene_data)  # update scene data

        return self

    def reset(self) -> OptProp:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            OptProp feature.
        """
        # Reset vop template
        if self.vop_template_link is not None:
            self._vop_template = self.vop_template_link.get()

        # Reset sop template
        if self.sop_template_link is not None:
            self._sop_template = self.sop_template_link.get()

        # Reset material instance
        if self._project.scene_link is not None:
            scene_data = self._project.scene_link.get()  # retrieve scene data
            # Look if an element corresponds to the _unique_id
            mat_inst = next(
                (x for x in scene_data.materials if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if mat_inst is not None:
                self._material_instance = mat_inst
        return self

    def delete(self) -> OptProp:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.opt_prop.OptProp
            OptProp feature.
        """
        # Delete the vop template
        if self.vop_template_link is not None:
            self.vop_template_link.delete()
            self.vop_template_link = None

        # Reset then the vop_guid (as the vop template was deleted just above)
        self._material_instance.vop_guid = ""

        # Delete the sop template
        if self.sop_template_link is not None:
            self.sop_template_link.delete()
            self.sop_template_link = None

        # Reset then the sop_guids (as the sop template was deleted just above)
        self._material_instance.ClearField("sop_guids")

        # Remove the material instance from the scene
        scene_data = self._project.scene_link.get()  # retrieve scene data
        mat_inst = next(
            (x for x in scene_data.materials if x.metadata["UniqueId"] == self._unique_id),
            None,
        )
        if mat_inst is not None:
            scene_data.materials.remove(mat_inst)
            self._project.scene_link.set(data=scene_data)  # update scene data

        # Reset the _unique_id
        self._unique_id = None
        self._material_instance.metadata.pop("UniqueId")
        return self

    def _fill(self, mat_inst):
        self._unique_id = mat_inst.metadata["UniqueId"]
        self._material_instance = mat_inst
        self.vop_template_link = self._project.client[mat_inst.vop_guid]
        if len(mat_inst.sop_guids) > 0:
            self.sop_template_link = self._project.client[mat_inst.sop_guids[0]]
        else:  # Specific case for ambient material
            self._sop_template = None
        self.reset()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Import geometries and materials from several SPEOS files to a project."""

from pathlib import Path
from typing import List, Optional

from ansys.speos.core.kernel.part import PartLink, ProtoPart
from ansys.speos.core.project import Project
from ansys.speos.core.speos import Speos


class SpeosFileInstance:
    """Represents a SPEOS file containing geometries and materials.

    Geometries are placed in the root part of a project, and oriented according to the axis_system
    argument.

    Parameters
    ----------
    speos_file : str
        SPEOS file to be loaded.
    axis_system : Optional[List[float]]
        Location and orientation to define for the geometry of the SPEOS file,
        [Ox, Oy, Oz, Xx, Xy, Xz, Yx, Yy, Yz, Zx, Zy, Zz].
        By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.
    name : str
        Name chosen for the imported geometry. This name is used as subpart name under the root part
        of the project.
        By default, "" (meaning user has not defined a name), then the name of the SPEOS file
        without extension is taken.
        Note: Materials are named after the name. For instance name.material.1 representing the
        first material of the imported geometry.
    """

    def __init__(
        self,
        speos_file: str,
        axis_system: Optional[List[float]] = None,
        name: str = "",
    ) -> None:
        self.speos_file = speos_file
        """SPEOS file."""
        if axis_system is None:
            axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
        self.axis_system = axis_system
        """Location and orientation to define for the geometry of the SPEOS file."""
        self.name = name
        """Name for the imported geometry, and used to name the materials."""

        if self.name == "":
            self.name = Path(speos_file).stem


def insert_speos(project: Project, speos_to_insert: List[SpeosFileInstance]) -> None:
    """Import geometries and materials from the selected SPEOS files to the existing project.

    Geometries and materials are placed in the root part, and orientated thanks to the
    SpeosFileInstance object.

    Notes
    -----
    Sources, Sensors and Simulations are not imported to the project.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project in which to import geometries and materials from SPEOS files.
    speos_to_combine : List[ansys.speos.core.workflow.combine_speos.SpeosFileInstance]
        List of SPEOS files, location and orientation of geometries to be imported to the project.
    """
    # Part link : either create it empty if none is present in the project's scene
    # or just retrieve it from project's scene
    part_link = None
    if project.scene_link.get().part_guid == "":
        part_link = project.client.parts().create(message=ProtoPart())
    else:
        part_link = project.client[project.scene_link.get().part_guid]

    # Combine all speos_to_insert into the project
    _combine(project=project, part_link=part_link, speos_to_combine=speos_to_insert)


def combine_speos(speos: Speos, speos_to_combine: List[SpeosFileInstance]) -> Project:
    """Create a project by combining geometries and materials from the selected SPEOS files.

    Geometries and materials are placed in the root part,
    and orientated thanks to the SpeosFileInstance object.

    Notes
    -----
        Sources, Sensors and Simulations are not imported to the project.

    Parameters
    ----------
    speos : ansys.speos.core.speos.Speos
        Speos session (connected to gRPC server).
    speos_to_combine : List[ansys.speos.core.workflow.combine_speos.SpeosFileInstance]
        List of SPEOS files, location and orientation of geometries to be imported to the project.

    Returns
    -------
    ansys.speos.core.project.Project
        Project created by combining the input list of SPEOS files.
    """
    # Create an empty project and an empty part link
    p = Project(speos=speos)
    part_link = speos.client.parts().create(message=ProtoPart())

    # Combine all speos_to_combine into the project
    _combine(project=p, part_link=part_link, speos_to_combine=speos_to_combine)

    return p


def _combine(
    project: Project,
    part_link: PartLink,
    speos_to_combine: List[SpeosFileInstance],
):
    scene_data = project.scene_link.get()
    part_data = part_link.get()

    for spc in speos_to_combine:
        scene_tmp = project.client.scenes().create()
        scene_tmp.load_file(file_uri=spc.speos_file)
        scene_tmp_data = scene_tmp.get()

        part_inst = ProtoPart.PartInstance(name=spc.name)
        part_inst.axis_system[:] = spc.axis_system
        part_inst.part_guid = scene_tmp_data.part_guid
        part_data.parts.append(part_inst)

        for mat in scene_tmp_data.materials:
            if len(mat.sop_guids) > 0:
                mat.name = spc.name + "." + mat.name
                mat.geometries.geo_paths[:] = [spc.name + "/" + x for x in mat.geometries.geo_paths]
                scene_data.materials.append(mat)

    part_link.set(data=part_data)
    scene_data.part_guid = part_link.key
    project.scene_link.set(data=scene_data)

    project._fill_features()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a wrapped abstraction of the gRPC proto API definition and stubs."""

from typing import List

from ansys.api.speos.part.v1 import (
    part_pb2 as messages,
    part_pb2_grpc as service,
)
from ansys.speos.core.kernel.crud import CrudItem, CrudStub
from ansys.speos.core.kernel.proto_message_utils import protobuf_message_to_str

ProtoPart = messages.Part
"""Part protobuf class : ansys.api.speos.part.v1.part_pb2.Part"""
ProtoPart.__str__ = lambda self: protobuf_message_to_str(self)


class PartLink(CrudItem):
    """Link object for a part in database.

    Parameters
    ----------
    db : ansys.speos.core.kernel.part.PartStub
        Database to link to.
    key : str
        Key of the part in the database.
    """

    def __init__(self, db, key: str):
        super().__init__(db, key)

    def __str__(self) -> str:
        """Return the string representation of the part."""
        return str(self.get())

    def get(self) -> ProtoPart:
        """Get the datamodel from database.

        Returns
        -------
        part.Part
            Part datamodel.
        """
        return self._stub.read(self)

    def set(self, data: ProtoPart) -> None:
        """Change datamodel in database.

        Parameters
        ----------
        data : part.Part
            New part datamodel.
        """
        self._stub.update(self, data)

    def delete(self) -> None:
        """Remove datamodel from database."""
        self._stub.delete(self)


class PartStub(CrudStub):
    """
    Database interactions for part.

    Parameters
    ----------
    channel : grpc.Channel
        Channel to use for the stub.

    Examples
    --------
    The best way to get a PartStub is to retrieve it from SpeosClient via parts() method.
    Like in the following example:

    >>> from ansys.speos.core.speos import Speos
    >>> speos = Speos(host="localhost", port=50098)
    >>> part_db = speos.client.parts()

    """

    def __init__(self, channel):
        super().__init__(stub=service.PartsManagerStub(channel=channel))

    def create(self, message: ProtoPart) -> PartLink:
        """Create a new entry.

        Parameters
        ----------
        message : part.Part
            Datamodel for the new entry.

        Returns
        -------
        ansys.speos.core.kernel.part.PartLink
            Link object created.
        """
        resp = CrudStub.create(self, messages.Create_Request(part=message))
        return PartLink(self, resp.guid)

    def read(self, ref: PartLink) -> ProtoPart:
        """Get an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.part.PartLink
            Link object to read.

        Returns
        -------
        part.Part
            Datamodel of the entry.
        """
        if not ref.stub == self:
            raise ValueError("PartLink is not on current database")
        resp = CrudStub.read(self, messages.Read_Request(guid=ref.key))
        return resp.part

    def update(self, ref: PartLink, data: ProtoPart):
        """Change an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.part.PartLink
            Link object to update.

        data : part.Part
            New datamodel for the entry.
        """
        if not ref.stub == self:
            raise ValueError("PartLink is not on current database")
        CrudStub.update(self, messages.Update_Request(guid=ref.key, part=data))

    def delete(self, ref: PartLink) -> None:
        """Remove an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.part.PartLink
            Link object to delete.
        """
        if not ref.stub == self:
            raise ValueError("PartLink is not on current database")
        CrudStub.delete(self, messages.Delete_Request(guid=ref.key))

    def list(self) -> List[PartLink]:
        """List existing entries.

        Returns
        -------
        List[ansys.speos.core.kernel.part.PartLink]
            Link objects.
        """
        guids = CrudStub.list(self, messages.List_Request()).guids
        return list(map(lambda x: PartLink(self, x), guids))



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a wrapped abstraction of the gRPC proto API definition and stubs."""

from typing import Iterator, List

from ansys.api.speos.job.v2 import job_pb2 as messages, job_pb2_grpc as service
from ansys.api.speos.results.v1.ray_path_pb2 import RayPath
from ansys.speos.core.kernel.crud import CrudItem, CrudStub
from ansys.speos.core.kernel.proto_message_utils import protobuf_message_to_str

ProtoJob = messages.Job
"""Job protobuf class : ansys.api.speos.job.v2.job_pb2.Job"""
ProtoJob.__str__ = lambda self: protobuf_message_to_str(self)


class JobLink(CrudItem):
    """Link object for job in database.

    Parameters
    ----------
    db : ansys.speos.core.kernel.job.JobStub
        Database to link to.
    key : str
        Key of the job in the database.
    """

    def __init__(self, db, key: str):
        super().__init__(db, key)
        self._actions_stub = db._actions_stub

    def __str__(self) -> str:
        """Return the string representation of the Job."""
        return str(self.get())

    def get(self) -> ProtoJob:
        """Get the datamodel from database.

        Returns
        -------
        job.Job
            Job datamodel.
        """
        return self._stub.read(self)

    def set(self, data: ProtoJob) -> None:
        """Change datamodel in database.

        Parameters
        ----------
        data : job.Job
            New Job datamodel.
        """
        self._stub.update(self, data)

    def delete(self) -> None:
        """Remove datamodel from database."""
        self._stub.delete(self)

    # Actions
    def get_state(self) -> messages.GetState_Response:
        """
        Retrieve job state.

        Returns
        -------
        ansys.api.speos.job.v2.job_pb2.GetState_Response
            State of the job.
        """
        return self._actions_stub.GetState(messages.GetState_Request(guid=self.key))

    def start(self) -> None:
        """Start the job."""
        self._actions_stub.Start(messages.Start_Request(guid=self.key))

    def stop(self) -> None:
        """Stop the job."""
        self._actions_stub.Stop(messages.Stop_Request(guid=self.key))

    def get_error(self) -> messages.GetError_Response:
        """
        Retrieve job error.

        Returns
        -------
        ansys.api.speos.job.v2.job_pb2.GetError_Response
            Error of the job.
        """
        return self._actions_stub.GetError(messages.GetError_Request(guid=self.key))

    def get_results(self) -> messages.GetResults_Response:
        """
        Retrieve job results.

        Returns
        -------
        ansys.api.speos.job.v2.job_pb2.GetResults_Response
            Results of the job.
        """
        return self._actions_stub.GetResults(messages.GetResults_Request(guid=self.key))

    def get_progress_status(self) -> messages.GetProgressStatus_Response:
        """
        Retrieve job progress.

        Returns
        -------
        ansys.api.speos.job.v2.job_pb2.GetProgressStatus_Response
            Progress status of the job.
        """
        return self._actions_stub.GetProgressStatus(
            messages.GetProgressStatus_Request(guid=self.key)
        )

    def get_ray_paths(self) -> Iterator[RayPath]:
        """Retrieve ray paths.

        Available for interactive simulation.

        Returns
        -------
        Iterator[ansys.api.speos.results.v1.ray_path_pb2.RayPath]
            Ray paths generated by the interactive simulation.
        """
        for rp in self._actions_stub.GetRayPaths(messages.GetRayPaths_Request(guid=self.key)):
            yield rp


class JobStub(CrudStub):
    """
    Database interactions for job.

    Parameters
    ----------
    channel : grpc.Channel
        Channel to use for the stub.

    Examples
    --------
    The best way to get a JobStub is to retrieve it from SpeosClient via jobs() method.
    Like in the following example:

    >>> from ansys.speos.core.speos import Speos
    >>> speos = Speos(host="localhost", port=50098)
    >>> job_db = speos.client.jobs()

    """

    def __init__(self, channel):
        super().__init__(stub=service.JobsManagerStub(channel=channel))
        self._actions_stub = service.JobActionsStub(channel=channel)

    def create(self, message: ProtoJob) -> JobLink:
        """Create a new entry.

        Parameters
        ----------
        message : job.Job
            Datamodel for the new entry.

        Returns
        -------
        ansys.speos.core.kernel.job.JobLink
            Link object created.
        """
        resp = CrudStub.create(self, messages.Create_Request(job=message))
        return JobLink(self, resp.guid)

    def read(self, ref: JobLink) -> ProtoJob:
        """Get an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.job.JobLink
            Link object to read.

        Returns
        -------
        job.Job
            Datamodel of the entry.
        """
        if not ref.stub == self:
            raise ValueError("JobLink is not on current database")
        resp = CrudStub.read(self, messages.Read_Request(guid=ref.key))
        return resp.job

    def update(self, ref: JobLink, data: ProtoJob):
        """Change an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.job.JobLink
            Link object to update.

        data : job.Job
            New datamodel for the entry.
        """
        if not ref.stub == self:
            raise ValueError("JobLink is not on current database")
        CrudStub.update(self, messages.Update_Request(guid=ref.key, job=data))

    def delete(self, ref: JobLink) -> None:
        """Remove an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.job.JobLink
            Link object to delete.
        """
        if not ref.stub == self:
            raise ValueError("JobLink is not on current database")
        CrudStub.delete(self, messages.Delete_Request(guid=ref.key))

    def list(self) -> List[JobLink]:
        """List existing entries.

        Returns
        -------
        List[ansys.speos.core.kernel.job.JobLink]
            Link objects.
        """
        guids = CrudStub.list(self, messages.List_Request()).guids
        return list(map(lambda x: JobLink(self, x), guids))



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with feature: Part."""

from __future__ import annotations

import re
from typing import List, Mapping, Optional, Union
import uuid

from ansys.speos.core import proto_message_utils
import ansys.speos.core.body as body
import ansys.speos.core.face as face
from ansys.speos.core.geo_ref import GeoRef
from ansys.speos.core.kernel.client import SpeosClient
from ansys.speos.core.kernel.part import ProtoPart
import ansys.speos.core.project as project


class Part:
    """Feature : Part.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.

    Attributes
    ----------
    part_link : ansys.speos.core.kernel.part.PartLink
        Link object for the part in database.
    """

    class SubPart:
        """Feature : SubPart.

        Parameters
        ----------
        speos_client : ansys.speos.core.kernel.client.SpeosClient
            The Speos instance client.
        name : str
            Name of the feature.
        description : str
            Description of the feature.
            By default, ``""``.
        parent_part : ansys.speos.core.part.Part, optional
            Part containing this sub part.
            By default, ``None``.

        Attributes
        ----------
        part_link : ansys.speos.core.kernel.part.PartLink
            Link object for the part in database.
        """

        def __init__(
            self,
            speos_client: SpeosClient,
            name: str,
            description: str = "",
            parent_part: Optional[Part] = None,
        ) -> None:
            self._speos_client = speos_client
            self._parent_part = parent_part
            self._name = name
            self.part_link = None
            """Link object for the part in database."""
            self._unique_id = None
            self._part_instance = ProtoPart.PartInstance(name=name, description=description)

            # Create local Part
            self._part = ProtoPart(name=name, description=description)

            self._geom_features = []

        @property
        def geo_path(self) -> GeoRef:
            """Geometry path to be used within other speos objects."""
            geo_paths = [self._name]
            if isinstance(self._parent_part, Part.SubPart):
                geo_paths.insert(0, self._parent_part.geo_path.metadata["GeoPath"])
            return GeoRef.from_native_link("/".join(geo_paths))

        def create_body(
            self,
            name: str,
            description: str = "",
            metadata: Optional[Mapping[str, str]] = None,
        ) -> body.Body:
            """Create a body in this element.

            Parameters
            ----------
            name : str
                Name of the feature.
            description : str
                Description of the feature.
                By default, ``""``.
            metadata : Optional[Mapping[str, str]]
                Metadata of the feature.
                By default, ``{}``.

            Returns
            -------
            ansys.speos.core.body.Body
                Body feature.
            """
            if metadata is None:
                metadata = {}

            body_feat = body.Body(
                speos_client=self._speos_client,
                name=name,
                description=description,
                metadata=metadata,
                parent_part=self,
            )
            self._geom_features.append(body_feat)
            return body_feat

        def create_sub_part(self, name: str, description: str = "") -> Part.SubPart:
            """Create a sub part in this element.

            Parameters
            ----------
            name : str
                Name of the feature.
            description : str
                Description of the feature.
                By default, ``""``.

            Returns
            -------
            ansys.speos.core.part.Part.SubPart
                SubPart feature.
            """
            sub_part_feat = Part.SubPart(
                speos_client=self._speos_client,
                name=name,
                description=description,
                parent_part=self,
            )
            self._geom_features.append(sub_part_feat)
            return sub_part_feat

        def set_axis_system(self, axis_system: List[float]) -> Part.SubPart:
            """Set the sub part orientation (relatively to parent element).

            Parameters
            ----------
            axis_system : List[float], optional
                Orientation of the sub part [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].

            Returns
            -------
            ansys.speos.core.part.Part.SubPart
                SubPart feature.
            """
            self._part_instance.axis_system[:] = axis_system
            return self

        def _to_dict(self) -> dict:
            out_dict = ""

            # Part Instance
            if self._parent_part is not None and self._parent_part.part_link is not None:
                parent_part_data = self._parent_part.part_link.get()
                part_inst = next(
                    (
                        x
                        for x in parent_part_data.parts
                        if x.description == "UniqueId_" + self._unique_id
                    ),
                    None,
                )
                if part_inst is not None:
                    out_dict = proto_message_utils._replace_guids(
                        speos_client=self._speos_client, message=part_inst
                    )
                else:
                    out_dict = proto_message_utils._replace_guids(
                        speos_client=self._speos_client,
                        message=self._part_instance,
                    )
            else:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._speos_client, message=self._part_instance
                )

            if "part" not in out_dict.keys():
                # Part
                if self.part_link is None:
                    out_dict["part"] = proto_message_utils._replace_guids(
                        speos_client=self._speos_client, message=self._part
                    )
                else:
                    out_dict["part"] = proto_message_utils._replace_guids(
                        speos_client=self._speos_client,
                        message=self.part_link.get(),
                    )

            return out_dict

        def __str__(self) -> str:
            """Return the string representation of the sub part."""
            out_str = ""

            if self._parent_part is not None and self._parent_part.part_link is not None:
                parent_part_data = self._parent_part.part_link.get()
                part_inst = next(
                    (
                        x
                        for x in parent_part_data.parts
                        if x.description == "UniqueId_" + self._unique_id
                    ),
                    None,
                )
                if part_inst is None:
                    out_str += "local: "
            else:
                out_str += "local: "

            out_str += proto_message_utils.dict_to_str(dict=self._to_dict())
            return out_str

        def commit(self) -> Part.SubPart:
            """Save feature: send the local data to the speos server database.

            Returns
            -------
            ansys.speos.core.part.Part.SubPart
                SubPart feature.
            """
            # The _unique_id will help to find correct item in the scene.materials:
            # the list of MaterialInstance
            if self._unique_id is None:
                self._unique_id = str(uuid.uuid4())
                self._part_instance.description = "UniqueId_" + self._unique_id

            # Save or Update the part (depending on if it was already saved before)
            if self.part_link is None:
                self.part_link = self._speos_client.parts().create(message=self._part)
            elif self.part_link.get() != self._part:
                self.part_link.set(data=self._part)  # Only update if data has changed

            self._part_instance.part_guid = self.part_link.key

            for g in self._geom_features:
                g.commit()

            # Look if an element corresponds to the instance
            if self._parent_part is not None and self._parent_part.part_link is not None:
                update_part = True
                parent_part_data = self._parent_part.part_link.get()

                part_inst = next(
                    (
                        x
                        for x in parent_part_data.parts
                        if x.description == "UniqueId_" + self._unique_id
                    ),
                    None,
                )
                if part_inst is not None:
                    if part_inst != self._part_instance:
                        part_inst.CopyFrom(self._part_instance)  # if yes, just replace
                    else:
                        update_part = False
                else:
                    self._parent_part._part.parts.append(
                        self._part_instance
                    )  # if no, just add it to the list of part instances

                if self._parent_part.part_link is not None and update_part:
                    self._parent_part.part_link.set(
                        data=self._parent_part._part
                    )  # update parent part

            return self

        def reset(self) -> Part.SubPart:
            """Reset feature: override local data by the one from the speos server database.

            Returns
            -------
            ansys.speos.core.part.Part.SubPart
                SubPart feature.
            """
            # Reset part
            if self.part_link is not None:
                self._part = self.part_link.get()

            # Reset part instance
            if self._parent_part is not None and self._parent_part.part_link is not None:
                parent_part_data = self._parent_part.part_link.get()  # retrieve server data
                # Look if an element corresponds to the _unique_id
                if self._unique_id is not None:
                    part_inst = next(
                        (
                            x
                            for x in parent_part_data.parts
                            if x.description == "UniqueId_" + self._unique_id
                        ),
                        None,
                    )
                    if part_inst is not None:
                        self._part_instance = part_inst

            return self

        def delete(self) -> Part.SubPart:
            """Delete feature: delete data from the speos server database.

            Returns
            -------
            ansys.speos.core.part.Part.SubPart
                SubPart feature.
            """
            # Retrieve all features to delete them
            while len(self._geom_features) > 0:
                self._geom_features[0].delete()

            # Remove the part instance from the parent part
            if self._parent_part is not None and self._parent_part.part_link is not None:
                parent_part_data = self._parent_part.part_link.get()
                part_inst = next(
                    (
                        x
                        for x in parent_part_data.parts
                        if x.description == "UniqueId_" + self._unique_id
                    ),
                    None,
                )
                if part_inst is not None:
                    self._parent_part._part.parts.remove(part_inst)
                    if self._parent_part.part_link is not None:
                        self._parent_part.part_link.set(
                            data=self._parent_part._part
                        )  # update parent part

            # Reset the _unique_id
            self._unique_id = None
            self._part_instance.description = ""

            if self in self._parent_part._geom_features:
                self._parent_part._geom_features.remove(self)

            return self

        def find(
            self,
            name: str,
            name_regex: bool = False,
            feature_type: Optional[type] = None,
        ) -> List[Union[body.Body, face.Face, Part.SubPart]]:
            """Find feature(s).

            Parameters
            ----------
            name : str
                Name of the feature.
                Possibility to look also for bodies, faces, subpart.
                Example "BodyName/FaceName", "SubPartName/BodyName/FaceName"
            name_regex : bool
                Allows to use regex for name parameter.
                By default, ``False``, means that regex is not used for name parameter.
            feature_type : type
                Type of the wanted feature (example: ansys.speos.core.body.Body,\
                ansys.speos.core.face.Face, ansys.speos.core.part.Part.SubPart).
                By default, ``None``, means that all features will be considered.

            Returns
            -------
            List[Union[ansys.speos.core.body.Body, ansys.speos.core.face.Face,\
            ansys.speos.core.part.Part.SubPart]]
                Found features.
            """
            orig_name = name
            idx = name.find("/")
            if idx != -1:
                name = name[0:idx]

            found_features = []

            if idx == -1 and feature_type is not None:
                if name_regex:
                    p = re.compile(name)
                    found_features.extend(
                        [
                            x
                            for x in self._geom_features
                            if p.match(x._name) and isinstance(x, feature_type)
                        ]
                    )
                else:
                    found_features.extend(
                        [
                            x
                            for x in self._geom_features
                            if x._name == name and isinstance(x, feature_type)
                        ]
                    )
            else:
                if name_regex:
                    p = re.compile(name)
                    found_features.extend([x for x in self._geom_features if p.match(x._name)])
                else:
                    found_features.extend([x for x in self._geom_features if x._name == name])

            if found_features and idx != -1:
                tmp = [
                    f.find(
                        name=orig_name[idx + 1 :],
                        name_regex=name_regex,
                        feature_type=feature_type,
                    )
                    for f in found_features
                ]

                found_features.clear()
                for feats in tmp:
                    found_features.extend(feats)

            return found_features

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
    ) -> None:
        self._project = project
        self._name = name
        self.part_link = None
        """Link object for the part in database."""

        self._geom_features = []

        # Create local Part
        if metadata is None:
            metadata = {}
        self._part = ProtoPart(name=name, description=description, metadata=metadata)

    def create_body(
        self,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
    ) -> body.Body:
        """Create a body in this element.

        Parameters
        ----------
        name : str
            Name of the feature.
        description : str
            Description of the feature.
            By default, ``""``.
        metadata : Mapping[str, str]
            Metadata of the feature.
            By default, ``{}``.

        Returns
        -------
        ansys.speos.core.body.Body
            Body feature.
        """
        if metadata is None:
            metadata = {}

        body_feat = body.Body(
            speos_client=self._project.client,
            name=name,
            description=description,
            metadata=metadata,
            parent_part=self,
        )
        self._geom_features.append(body_feat)
        return body_feat

    def create_sub_part(self, name: str, description: str = "") -> Part.SubPart:
        """Create a sub part in this element.

        Parameters
        ----------
        name : str
            Name of the feature.
        description : str
            Description of the feature.
            By default, ``""``.

        Returns
        -------
        ansys.speos.core.part.Part.SubPart
            SubPart feature.
        """
        sub_part_feat = Part.SubPart(
            speos_client=self._project.client,
            name=name,
            description=description,
            parent_part=self,
        )
        self._geom_features.append(sub_part_feat)
        return sub_part_feat

    def _to_dict(self) -> dict:
        out_dict = ""

        if self.part_link is None:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._project.client, message=self._part
            )
        else:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._project.client, message=self.part_link.get()
            )

        return out_dict

    def __str__(self) -> str:
        """Return the string representation of the part."""
        out_str = ""

        if self.part_link is None:
            out_str += "local: "

        out_str += proto_message_utils.dict_to_str(dict=self._to_dict())
        return out_str

    def commit(self) -> Part:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.part.Part
            Part feature.
        """
        # Save or Update the part (depending on if it was already saved before)
        if self.part_link is None:
            self.part_link = self._project.client.parts().create(message=self._part)
        elif self.part_link.get() != self._part:
            self.part_link.set(data=self._part)  # Only update if data has changed

        # Retrieve all features to commit them
        for g in self._geom_features:
            g.commit()

        # Update the scene with the part
        if self._project.scene_link:
            scene_data = self._project.scene_link.get()  # retrieve scene data
            if scene_data.part_guid != self.part_link.key:
                scene_data.part_guid = self.part_link.key
                self._project.scene_link.set(data=scene_data)  # update scene data

        return self

    def reset(self) -> Part:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.part.Part
            Part feature.
        """
        # Reset part
        if self.part_link is not None:
            self._part = self.part_link.get()

        return self

    def delete(self) -> Part:
        """Delete feature: delete data from the speos server database.

        Returns
        -------
        ansys.speos.core.part.Part
            Part feature.
        """
        # Retrieve all features to delete them
        while len(self._geom_features) > 0:
            self._geom_features[0].delete()

        # Delete the part
        if self.part_link is not None:
            self.part_link.delete()
            self.part_link = None

        # Remove the part guid from the scene
        scene_data = self._project.scene_link.get()  # retrieve scene data
        scene_data.part_guid = ""
        self._project.scene_link.set(data=scene_data)  # update scene data

        return self

    def find(
        self,
        name: str,
        name_regex: bool = False,
        feature_type: Optional[type] = None,
    ) -> List[Union[body.Body, face.Face, Part.SubPart]]:
        """Find feature(s).

        Parameters
        ----------
        name : str
            Name of the feature.
            Possibility to look also for bodies, faces, subpart.
            Example "BodyName/FaceName", "SubPartName/BodyName/FaceName"
        name_regex : bool
            Allows to use regex for name parameter.
            By default, ``False``, means that regex is not used for name parameter.
        feature_type : type
            Type of the wanted feature (example: ansys.speos.core.body.Body,\
            ansys.speos.core.face.Face, ansys.speos.core.part.Part.SubPart).
            By default, ``None``, means that all features will be considered.

        Returns
        -------
        List[Union[ansys.speos.core.body.Body, ansys.speos.core.face.Face,\
        ansys.speos.core.part.Part.SubPart]]
            Found features.
        """
        orig_name = name
        idx = name.find("/")
        if idx != -1:
            name = name[0:idx]

        found_features = []
        if idx == -1 and feature_type is not None:
            if name_regex:
                p = re.compile(name)
                found_features.extend(
                    [
                        x
                        for x in self._geom_features
                        if p.match(x._name) and isinstance(x, feature_type)
                    ]
                )
            else:
                found_features.extend(
                    [
                        x
                        for x in self._geom_features
                        if x._name == name and isinstance(x, feature_type)
                    ]
                )
        else:
            if name_regex:
                p = re.compile(name)
                found_features.extend([x for x in self._geom_features if p.match(x._name)])
            else:
                found_features.extend([x for x in self._geom_features if x._name == name])

        if found_features and idx != -1:
            tmp = [
                f.find(
                    name=orig_name[idx + 1 :],
                    name_regex=name_regex,
                    feature_type=feature_type,
                )
                for f in found_features
            ]

            found_features.clear()
            for feats in tmp:
                found_features.extend(feats)

        return found_features



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module with utility elements for protobuf messages from Speos RPC server."""

import json

from google.protobuf import __version__ as protobuf_version
from google.protobuf.json_format import MessageToJson
from google.protobuf.message import Message


def protobuf_message_to_str(message: Message, with_full_name: bool = True) -> str:
    """
    Convert a protobuf message to formatted json string.

    Parameters
    ----------
    message : google.protobuf.message.Message
        Protobuf message.
    with_full_name : bool
        Prepend the returned string with protobuf message full name.
        By default, ``True``.

    Returns
    -------
    str
        protobuf message formatted to be logged/printed.
    """
    ret = ""
    if with_full_name:
        ret += message.DESCRIPTOR.full_name + "\n"

    protobuf_major_version = int(protobuf_version.split(sep=".", maxsplit=1)[0])
    if protobuf_major_version < 5:
        ret += MessageToJson(
            message=message,
            including_default_value_fields=True,
            preserving_proto_field_name=True,
            indent=4,
        )
    else:
        ret += MessageToJson(
            message=message,
            always_print_fields_with_no_presence=True,
            preserving_proto_field_name=True,
            indent=4,
        )
    return ret


def protobuf_message_to_dict(message) -> dict:
    """Convert protobuf message to a formatted json dict.

    Parameters
    ----------
    message : google.protobuf.message.Message
        Protobuf message.

    Returns
    -------
    dict
        protobuf message formatted as dict.
    """
    return json.loads(protobuf_message_to_str(message=message, with_full_name=False))
